from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from datetime import datetime, timezone, timedelta
from urllib.parse import urlparse
from typing import Optional
import base64
import re

ALGORITHM = "AES-256-GCM"

def get_utc_now(extra_seconds: Optional[int] = 0) -> str:
    """
    Get the current date and time in UTC format as an ISO string.

    Returns:
        str: Current UTC datetime in ISO format with 'Z' suffix (e.g., '2024-01-15T10:30:45.123Z')
    """
    # return datetime.now(timezone.utc).isoformat(timespec='milliseconds').replace('+00:00', 'Z')
    now = datetime.now(timezone.utc) + timedelta(seconds=extra_seconds)
    return now.isoformat(timespec='milliseconds').replace('+00:00', 'Z')


def decrypt(ciphertext: str, encryption_key: str) -> str:
    """Decrypts a Base64-encoded ciphertext string using AES-256-GCM.

    Args:
        ciphertext (str): The Base64-encoded string containing IV, authentication tag, and ciphertext.

    Returns:
        str: The decrypted plaintext string.
    """
    key: bytes = bytes.fromhex(encryption_key)
    data: bytes = base64.b64decode(ciphertext)

    iv: bytes = data[:12]
    auth_tag: bytes = data[12:28]
    encrypted: bytes = data[28:]

    decryptor = Cipher(
        algorithms.AES(key),
        modes.GCM(iv, auth_tag),
        backend=default_backend()
    ).decryptor()

    decrypted: bytes = decryptor.update(encrypted) + decryptor.finalize()
    return decrypted.decode("utf-8")

def validate_id_format(value: str):
    if not re.match(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', value):
        raise ValueError('ID deve ser um UUID válido')
    version = int(value[14], 16)
    if version != 7:
        raise ValueError('ID deve ser um UUID7 válido')
    return value

def is_valid_url(url: str) -> bool:
    try:
        result = urlparse(url)
    except ValueError:
        return False

    return all([result.scheme in ("http", "https"), result.netloc])


from zsynctech_studio_sdk.core.context import config_client
from zsynctech_studio_sdk.core.execution import Execution
from zsynctech_studio_sdk.core.task import Task
from zsynctech_studio_sdk.core.step import Step

__all__ = [
    "Execution",
    "Task",
    "Step",
    "config_client",
]


from zsynctech_studio_sdk.enums.execution import ExecutionStatus
from pydantic import BaseModel, Field, field_validator
from zsynctech_studio_sdk.enums.step import StepStatus
from zsynctech_studio_sdk.enums.task import TaskStatus
from uuid_extensions.uuid7 import uuid7
from typing import Optional, Union
from datetime import datetime
import re


class BaseEntity(BaseModel):
    id: str = Field(
        default_factory=lambda: str(uuid7()),
        min_length=1,
        description="ID único da execução"
    )
    observation: Optional[str] = None
    status: Optional[Union[ExecutionStatus, StepStatus, TaskStatus]] = None
    endDate: Optional[str] = None

    @field_validator('id')
    @classmethod
    def validate_id_format(cls, v):
        if not re.match(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', v):
            raise ValueError('ID deve ser um UUID válido')
        version = int(v[14], 16)
        if version != 7:
            raise ValueError('ID deve ser um UUID7 válido')
        return v

    @field_validator('endDate')
    @classmethod
    def validate_end_date_format(cls, v):
        if v is None:
            return v
        pattern = r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}Z$'
        if not re.match(pattern, v):
            raise ValueError('endDate deve estar no formato ISO 8601 com Z')
        try:
            datetime.fromisoformat(v.replace('Z', '+00:00'))
        except ValueError:
            raise ValueError('endDate deve ser uma data válida')
        return v

    class Config:
        extra = "forbid"
        validate_assignment = True


from zsynctech_studio_sdk.utils import validate_id_format, is_valid_url
from contextvars import ContextVar
from appdirs import AppDirs
from httpx import Client
import os

APPS_DIR = AppDirs()
INSTANCE_ID: ContextVar[str] = ContextVar("instance_id", default=None)
SERVER: ContextVar[str] = ContextVar("server", default=None)
ENCRYPTION_KEY: ContextVar[str] = ContextVar("encryption_key", default=None)
SECRET_KEY: ContextVar[str] = ContextVar("secret_key", default=None)
CLIENT: ContextVar[object] = ContextVar("client", default=None)
SDK_DIR = os.path.join(APPS_DIR.user_data_dir, "zsynctech")


def config_client(server: str, secret_key: str, instance_id: str, encryption_key: str) -> Client:
    """Sets the credentials for the SDK.

    Args:
        server (str): The server URL.
        secret_key (str): The secret key.
        instance_id (str): The instance ID.
        encryption_key (str): The encryption key.

    Returns:
        Client: The HTTP client configured with the provided credentials.
    """
    try:
        validate_id_format(instance_id)
    except Exception:
        raise RuntimeError("Invalid instance ID")

    try:
        validate_id_format(secret_key)
    except Exception:
        raise RuntimeError("Invalid secret key")

    if not is_valid_url(server):
        raise RuntimeError("Invalid server url")

    INSTANCE_ID.set(instance_id)
    ENCRYPTION_KEY.set(encryption_key)
    SERVER.set(server.rstrip('/'))
    SECRET_KEY.set(secret_key)

    base_url = f"{SERVER.get()}/automation-gateway"

    if CLIENT.get() is None:
        client = Client(
            base_url=base_url,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {secret_key}::{instance_id}"
            }
        )
        CLIENT.set(client)
    
    return CLIENT.get()

def get_client() -> Client:
    """Gets the HTTP client.

    Raises:
        RuntimeError: If the client is not initialized.

    Returns:
        Client: The HTTP client.
    """
    if CLIENT.get() is None:
        raise RuntimeError("HTTP client not initialized. Call set_credentials() first.")
    return CLIENT.get()


from loguru import logger as _base_logger
import sys


def get_logger(name: str = "zsynctech-studio-sdk"):
    """_summary_

    Args:
        name (str, optional): _description_. Defaults to "zsynctech-studio-sdk".

    Returns:
        _type_: _description_
    """
    logger = _base_logger.bind(library=name)
    logger.remove()

    logger.add(
        sys.stdout,
        colorize=True,
        level="DEBUG",
        format=(
            "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
            "<level>{level: <8}</level> | "
            "<cyan>{extra[library]}</cyan> | "
            "<magenta>{module}</magenta>:<yellow>{name}</yellow>:<red>{line}</red> | "
            "<level>{message}</level>"
        ),
    )

    return logger


from zsynctech_studio_sdk.models.base import BaseEntity
from zsynctech_studio_sdk.enums.step import StepStatus
from pydantic import field_validator
from datetime import datetime
from typing import Optional
import re


class StepModel(BaseEntity):
    status: StepStatus = StepStatus.UNPROCESSED
    taskId: Optional[str] = None
    stepCode: Optional[str] = None
    startDate: Optional[str] = None
    automationOnClientId: str

    @field_validator('startDate')
    @classmethod
    def validate_end_date_format(cls, v):
        if v is None:
            return v
        pattern = r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}Z$'
        if not re.match(pattern, v):
            raise ValueError('endDate deve estar no formato ISO 8601 com Z')
        try:
            datetime.fromisoformat(v.replace('Z', '+00:00'))
        except ValueError:
            raise ValueError('endDate deve ser uma data válida')
        return v



from zsynctech_studio_sdk.enums.execution import ExecutionStatus
from zsynctech_studio_sdk.models.base import BaseEntity
from pydantic import model_validator


class ExecutionModel(BaseEntity):
    totalTaskCount: int = 0
    currentTaskCount: int = 0
    status: ExecutionStatus = ExecutionStatus.WAITING

    @model_validator(mode='after')
    def validate_business_rules(self):
        current = self.currentTaskCount or 0
        total = self.totalTaskCount or 0
        if current > total:
            raise ValueError('currentTaskCount não pode ser maior que totalTaskCount')

        return self



from typing import Optional, List
from pydantic import BaseModel
from enum import StrEnum


class InputOutputTypes(StrEnum):
    FTP = 'FTP'
    API = 'API'
    QUEUE = 'QUEUE'


class CredentialModel(BaseModel):
    key: str
    value: str
    encrypted: bool


class ExecutionSettings(BaseModel):
    instanceId: str
    executionId: str
    automationName: Optional[str] = "System"
    clientId: Optional[str] = None
    userId: Optional[str] = "System"
    outputPath: Optional[str] = None
    inputPath: Optional[str] = None
    inputMetaData: Optional[dict] = None
    inputType: Optional[InputOutputTypes] = InputOutputTypes.FTP
    outputType: Optional[InputOutputTypes] = InputOutputTypes.FTP
    outputMetaData: Optional[dict] = None
    keepAlive: Optional[bool] = False
    keepAliveInterval: Optional[int] = 30
    credentials: Optional[List[CredentialModel]] = None



from zsynctech_studio_sdk.core.context import get_client, INSTANCE_ID
from zsynctech_studio_sdk.core.loggers import get_logger
from zsynctech_studio_sdk.models.step import StepModel
from zsynctech_studio_sdk.enums.step import StepStatus
from zsynctech_studio_sdk.utils import get_utc_now
from typing import Optional
from functools import wraps


STEP_FINALIZED_STATUS = [
    StepStatus.SUCCESS,
    StepStatus.FAIL
]

logger = get_logger()

def send_step(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        result = method(self, *args, **kwargs)
        if self.model.status in STEP_FINALIZED_STATUS:
            self.model.endDate = get_utc_now(1)
        self._client.post(
            url="taskSteps",
            json=self.model.model_dump()
        ).raise_for_status()
        return result
    return wrapper


class Step:
    def __init__(self, step_code: str, task_id: str):
        self.model = StepModel(
            stepCode=step_code,
            automationOnClientId=INSTANCE_ID.get(),
            taskId=task_id,
            status=StepStatus.UNPROCESSED
        )
        self._client = get_client()

    @send_step
    def start(self, observation: Optional[str] = "Step iniciado") -> 'Step':
        """Mark the step as started.

        Args:
            observation (Optional[str], optional): Observation text. Defaults to "Step iniciado".

        Returns:
            Step: The current step instance.
        """
        self.model.status = StepStatus.RUNNING
        self.model.startDate = get_utc_now()
        self.model.observation = observation
        logger.info(f"Step code: {self.model.stepCode} - Observation: {self.model.observation}")
        return self

    @send_step
    def success(self, observation: Optional[str] = "Step finalizado com sucesso") -> 'Step':
        """Mark the step as success.

        Args:
            observation (Optional[str], optional): Observation text. Defaults to "Step iniciado".

        Returns:
            Step: The current step instance.
        """
        self.model.status = StepStatus.SUCCESS
        self.model.observation = observation
        logger.success(f"Step code: {self.model.stepCode} - Observation: {self.model.observation}")
        return self

    @send_step
    def error(self, observation: Optional[str] = "Ocorreu um erro no step") -> 'Step':
        """Mark the step as fail.

        Args:
            observation (Optional[str], optional): Observation text. Defaults to "Ocorreu um erro no step".

        Returns:
            Step: The current step instance.
        """
        self.model.status = StepStatus.FAIL
        self.model.observation = observation
        logger.error(f"Step code: {self.model.stepCode} - Observation: {self.model.observation}")
        return self


from zsynctech_studio_sdk.core.context import get_client, INSTANCE_ID, SDK_DIR, ENCRYPTION_KEY
from zsynctech_studio_sdk.models.config import ExecutionSettings, CredentialModel
from zsynctech_studio_sdk.models.execution import ExecutionModel
from zsynctech_studio_sdk.enums.execution import ExecutionStatus
from zsynctech_studio_sdk.core.loggers import get_logger
from zsynctech_studio_sdk.utils import get_utc_now
from zsynctech_studio_sdk.core.task import Task
from zsynctech_studio_sdk.utils import decrypt
from typing import Optional, Dict, List
from cachelib import FileSystemCache
from uuid_extensions import uuid7
from functools import wraps
import json
import os

EXECUTION_FINISHED_STATUS = [
    ExecutionStatus.OUT_OF_OPERATING_HOURS,
    ExecutionStatus.INTERRUPTED,
    ExecutionStatus.FINISHED,
    ExecutionStatus.ERROR,
]

logger = get_logger()

def is_a_finished_execution(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self.model.status in EXECUTION_FINISHED_STATUS:
            logger.debug(
                f"Ignorando '{method.__name__}' — execução já finalizada com status '{self.model.status.value}'."
            )
            return self
        return method(self, *args, **kwargs)
    return wrapper

def send_execution(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self.model.status in EXECUTION_FINISHED_STATUS:
            logger.debug(
                f"Ignorando envio de '{method.__name__}' — execução já finalizada ({self.model.status.value})."
            )
            return self
        
        result = method(self, *args, **kwargs)

        if self.model.status in EXECUTION_FINISHED_STATUS:
            self.model.endDate = get_utc_now()

        try:
            self._client.post(
                url="executions",
                json=self.model.model_dump()
            ).raise_for_status()
        except Exception as e:
            logger.exception(f"Falha ao atualizar execução: {e}")

        return result
    return wrapper


class Execution:
    def __init__(self) -> None:
        self.model = ExecutionModel()
        self._client = get_client()
        self._cache = FileSystemCache(
            cache_dir=os.path.join(SDK_DIR, INSTANCE_ID.get()),
            default_timeout=0
        )
        self._settings: Optional[ExecutionSettings] = None
    
    @property
    def settings(self):
        return self._settings

    def task(
            self,
            code: Optional[str] = str(uuid7()),
            description: Optional[str] = "Não informado"
        ) -> Task:
        """Create a new task within the execution.

        Args:
            code (Optional[str], optional): Taks code. Defaults to str(uuid7()).
            description (Optional[str], optional): Text description. Defaults to "Não informado".

        Returns:
            Task: The created task instance.
        """
        task = Task(
            description,
            execution_id=self.model.id,
            code=code
        )
        return task

    @staticmethod
    def decrypt_credentials(credentials: Optional[List[CredentialModel]]) -> Optional[Dict[str, str]]:
        """Decrypts credentials and returns them as a dictionary {key: value}.

        Args:
            credentials (Optional[List[Credential]]): A list of Credential objects, or None.

        Returns:
            Optional[Dict[str, str]]: A dictionary with decrypted values, or None if no credentials are provided.
        """
        if credentials is None:
            return None

        decrypted_dict: Dict[str, str] = {}
        for cred in credentials:
            if cred.encrypted:
                decrypted_dict[cred.key] = decrypt(cred.value, ENCRYPTION_KEY.get())
            else:
                decrypted_dict[cred.key] = cred.value

        return decrypted_dict

    @send_execution
    @is_a_finished_execution
    def start(self, observation: Optional[str] = "Execução iniciada") -> 'Execution':
        """Mark the execution as started.

        Args:
            observation (Optional[str], optional): Observation text. Defaults to None.

        Returns:
            Execution: The current execution instance.
        """
        self.model.observation = observation
        self.model.status = ExecutionStatus.RUNNING
        logger.info(self.model.observation)
        return self

    @send_execution
    @is_a_finished_execution
    def finish(self, observation: Optional[str] = "Execução finalizada com sucesso") -> 'Execution':
        """Mark the execution as finish.

        Args:
            observation (Optional[str], optional): Observation text. Defaults to None.

        Returns:
            Execution: The current execution instance.
        """
        self.model.status = ExecutionStatus.FINISHED
        self.model.observation = observation
        logger.success(self.model.observation)
        return self

    @send_execution
    @is_a_finished_execution
    def out_of_operating_hours(self, observation: Optional[str] = "Fora do horário de execução") -> 'Execution':
        """Mark the execution as out of operating hours.

        Args:
            observation (Optional[str], optional): Observation text. Defaults to "Fora do horário de execução".

        Returns:
            Execution: The current execution instance.
        """
        self.model.status = ExecutionStatus.OUT_OF_OPERATING_HOURS
        self.model.observation = observation
        logger.error(self.model.observation)
        return self

    @send_execution
    @is_a_finished_execution
    def error(self, observation: Optional[str] = "Ocorreu um erro na execução") -> 'Execution':
        """Mark the execution as finish.

        Args:
            observation: str Observation text.

        Returns:
            Execution: The current execution instance.
        """
        self.model.status = ExecutionStatus.ERROR
        self.model.observation = observation
        logger.error(self.model.observation)
        return self

    @send_execution
    @is_a_finished_execution
    def interrupted(self, observation: Optional[str] = "Operação interrompida pelo sistema") -> 'Execution':
        """Mark the execution as interrupted.

        Args:
            observation: str Observation text.

        Returns:
            Execution: The current execution instance.
        """
        self.model.status = ExecutionStatus.OUT_OF_OPERATING_HOURS
        self.model.observation = observation
        logger.error(self.model.observation)
        return self

    @send_execution
    @is_a_finished_execution
    def waiting(self, observation: Optional[str] = "Execução em espera") -> 'Execution':
        """Mark the execution as waiting.

        Args:
            observation: str Observation text.

        Returns:
            Execution: The current execution instance.
        """
        self.model.status = ExecutionStatus.WAITING
        self.model.observation = observation
        logger.info(self.model.observation)
        return self

    @send_execution
    @is_a_finished_execution
    def scheduled(self, observation: Optional[str] = "Execução agendada para o próximo horário disponível") -> 'Execution':
        """Mark the execution as scheduled.

        Args:
            observation: str Observation text.

        Returns:
            Execution: The current execution instance.
        """
        self.model.status = ExecutionStatus.SCHEDULED
        self.model.observation = observation
        logger.info(self.model.observation)
        return self

    @send_execution
    def increment_task_count(self) -> 'Execution':
        """Increment the current task count by one.

        Args:
            observation: str Observation text.

        Returns:
            Execution: The current execution instance.
        """
        if self.model.totalTaskCount == 0 or self.model.currentTaskCount >= self.model.totalTaskCount:
            self.model.totalTaskCount += 1
        self.model.currentTaskCount = self.model.currentTaskCount + 1
        logger.info(f"Processado {self.model.currentTaskCount} de {self.model.totalTaskCount}")
        return self

    @send_execution
    def update_current_task_count(self, current_task_count: int) -> 'Execution':
        """Update the current task count to a specific value.

        Args:
            observation: str Observation text.

        Returns:
            Execution: The current execution instance.
        """
        self.model.currentTaskCount = current_task_count
        logger.info(f"Processado {self.model.currentTaskCount} de {self.model.totalTaskCount}")
        return self

    @send_execution
    def set_total_task_count(self, total_task_count: int) -> 'Execution':
        """Set the total task count to a specific value.

        Args:
            observation: str Observation text.

        Returns:
            Execution: The current execution instance.
        """
        self.model.totalTaskCount = total_task_count
        logger.info(f"Total de tasks a processar: {self.model.totalTaskCount}")
        return self

    def __get_latest_instance_file(self) -> Optional[str]:
        """Gets the latest instance file path based on modification time.
        Args:

        Returns:
            Optional[str]: The path to the latest instance file, or None if not found.
        """
        candidates = [
            os.path.join(SDK_DIR, f)
            for f in os.listdir(SDK_DIR)
            if f.endswith(".json") and f.startswith(INSTANCE_ID.get())
        ]
        if not candidates:
            return None
        return max(candidates, key=os.path.getmtime)

    def get_next_execution(self) -> Optional[ExecutionSettings]:
        """Gets the next execution settings from the latest instance file.

        Returns:
            Optional[ExecutionSettings]: Execution settings
        """
        cached = self._cache.get("settings")

        latest_file = self.__get_latest_instance_file()
        if not latest_file:
            logger.info("Sem atualizações")
            return None

        current_mtime = os.path.getmtime(latest_file)

        if cached and cached["path"] == latest_file and cached["mtime"] == current_mtime:
            logger.info("Sem atualizações")
            return None

        try:
            with open(latest_file, mode="r", encoding="utf8") as file:
                data = json.load(file)
        except Exception as e:
            logger.exception(e)
            return None

        self.model.id = data["executionId"]

        new_cache = {
            "path": latest_file,
            "mtime": current_mtime,
            "data": data
        }
        self._cache.set("settings", new_cache)

        self._settings = ExecutionSettings(**data)

        logger.info(f"Execução iniciada - {self._settings.executionId}")

        return self._settings

    def get_last_execution(self) -> Optional[ExecutionSettings]:
        """Gets the last execution settings from the cache.

        Returns:
            Optional[ExecutionSettings]: Execution settings
        """
        cached = self._cache.get("settings")
        if not cached:
            return None

        if "data" in cached:
            data = cached["data"]
            self.model.id = data["executionId"]
            self._settings = ExecutionSettings(**data)
            logger.info(f"Última execução (cache): {self._settings.executionId}")
            return self._settings

        try:
            with open(cached["path"], mode="r", encoding="utf8") as file:
                data = json.load(file)
        except Exception:
            return None

        self.model.id = data["executionId"]
        cached["data"] = data
        self._cache.set("settings", cached)

        self._settings = ExecutionSettings(**data)

        logger.info(f"Última execução: {self._settings.executionId}")

        return self._settings


from zsynctech_studio_sdk.enums.execution import ExecutionStatus
from zsynctech_studio_sdk.enums.task import TaskStatus
from zsynctech_studio_sdk.enums.step import StepStatus


__all__ = [
    "ExecutionStatus",
    "TaskStatus",
    "StepStatus",
]


from enum import StrEnum


class ExecutionStatus(StrEnum):
    WAITING = 'WAITING'
    RUNNING = 'RUNNING'
    FINISHED = 'FINISHED'
    ERROR = 'ERROR'
    SCHEDULED = 'SCHEDULED'
    INTERRUPTED = 'INTERRUPTED'
    OUT_OF_OPERATING_HOURS = 'OUT_OF_OPERATING_HOURS'



from zsynctech_studio_sdk.core.context import get_client
from zsynctech_studio_sdk.core.loggers import get_logger
from zsynctech_studio_sdk.models.task import TaskModel
from zsynctech_studio_sdk.enums.task import TaskStatus
from zsynctech_studio_sdk.utils import get_utc_now
from zsynctech_studio_sdk.core.step import Step
from typing import Dict

logger = get_logger()


class Task:
    def __init__(self, description: str, execution_id: str, code: str):
        self.model = TaskModel(
            description=description,
            executionId=execution_id,
            code=code
        )
        self.steps: Dict[str, Step] = {}
        self._client = get_client()

    def step(self, step_code: str) -> Step:
        """Create or retrieve a step within the task.

        Args:
            step_code (str): Step code.

        Returns:
            Step: The created or retrieved step instance.
        """
        step = self.steps.get(step_code)
        if step is None:
            step = Step(
                step_code,
                task_id=self.model.id
            )
        self.steps[step_code] = step
        return step
    

    def __enter__(self) -> 'Task':
        """Mark the task as started.

        Returns:
            Task: The current task instance.
        """
        self.model.status = TaskStatus.RUNNING
        self.model.startDate = get_utc_now()
        self._client.post(
            url="tasks",
            json=self.model.model_dump()
        ).raise_for_status()
        logger.info(f"Task code: {self.model.code} iniciada")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        """Mark the task as completed or failed based on the presence of an exception.

        Args:
            exc_type (_type_): _type_ of the exception, if any.
            exc_val (_type_): _value_ of the exception, if any.
            exc_tb (_type_): _traceback_ of the exception, if any.

        Returns:
            bool: False to propagate the exception, True to suppress it.
        """
        if exc_type:
            self.model.status = TaskStatus.FAIL
            self.model.observation = str(exc_val)
            logger.error(f"Ocorreu um erro ao executar a task - {self.model.code}")
        else:
            self.model.status = TaskStatus.SUCCESS

        self.model.endDate = get_utc_now()

        self._client.post(
            url="tasks",
            json=self.model.model_dump()
        ).raise_for_status()

        logger.success(f"Task code: {self.model.code} finalizada com sucesso")
        
        return False